"""Solver parameters

.. autofunction:: iter_complete_params

.. autoclass:: Parameters
   :members:
   :private-members:

"""

from pathlib import Path

from typing import Iterable

import h5py
from fluiddyn.util import import_class, mpi
from fluiddyn.util.paramcontainer import ParamContainer

from .info import InfoSolverCore


def iter_complete_params(params, info_solver, classes: dict or Iterable):
    """Iterate over a collection of classes and invoke the static method
    ``_complete_params_with_default`` to populate an instance of
    :class:`Parameters` in-place.

    Parameters
    ----------
    params : :class:`fluidsim_core.params.Parameters` or its subclass

    info_solver : :class:`fluidsim_core.info.InfoSolverCore` or its subclass

    classes : dict or iterable

    """

    if isinstance(classes, dict):
        classes = [classes[key] for key in sorted(classes.keys())]

    for Class in classes:
        if hasattr(Class, "_complete_params_with_default"):
            try:
                Class._complete_params_with_default(params)
            except TypeError:
                try:
                    Class._complete_params_with_default(params, info_solver)
                except TypeError as e:
                    e.args += ("for class: " + repr(Class),)
                    raise


class Parameters(ParamContainer):
    """Contain the parameters."""

    @classmethod
    def _create_params(cls, input_info_solver):
        """Create a Parameters instance from an InfoSolverCore instance."""
        if isinstance(input_info_solver, InfoSolverCore):
            info_solver = input_info_solver
        elif hasattr(input_info_solver, "Simul"):
            info_solver = input_info_solver.Simul.create_default_params()
        else:
            raise ValueError(
                "Can not create params from input input_info_solver."
            )

        params = cls(tag="params")

        dict_classes = {
            "Solver": import_class(
                info_solver.module_name, info_solver.class_name
            )
        }
        dict_classes.update(info_solver.import_classes())

        iter_complete_params(params, info_solver, dict_classes.values())
        return params

    @classmethod
    def _load_params_simul(cls, path=None, only_mpi_rank0=True):
        """Load the parameters and return a Parameters instance."""
        if mpi.rank > 0 and not only_mpi_rank0:
            params = None
        else:
            if path is None:
                path = Path.cwd()
            else:
                path = Path(path)

            path_xml = None
            if path.is_dir():
                path_xml = path / "params_simul.xml"
            elif path.name.endswith(".xml"):
                if not path.exists():
                    raise ValueError(f"The file {path} does not exists.")

                path_xml = path

            if path_xml is not None and path_xml.exists():
                params = cls(path_file=path_xml)
            else:
                if path.is_file():
                    paths = [path]
                else:
                    paths = list(path.glob("state_*"))
                if paths:
                    path = sorted(paths)[0]
                    str_path = str(path)
                    if len(str_path) > 100:
                        str_path = "[...]" + str_path[-100:]

                    print("Loading params from file\n" + str_path)
                    with h5py.File(path, "r") as h5file:
                        params = cls(hdf5_object=h5file["/info_simul/params"])
                else:
                    raise ValueError(f"No state files found in {path = }")

        if mpi.nb_proc > 1 and not only_mpi_rank0:
            params = mpi.comm.bcast(params, root=0)
        return params

    @classmethod
    def _load_info_solver(cls, path_dir=None):
        """Load the solver information, return an InfoSolverCore instance."""
        if path_dir is None:
            path_dir = Path.cwd()

        if not isinstance(path_dir, Path):
            path_dir = Path(path_dir)

        if not path_dir.is_dir():
            raise ValueError(str(path_dir) + " is not a directory")

        path_info_solver = path_dir / "info_solver.xml"
        if path_info_solver.exists():
            return cls(path_file=str(path_info_solver))

        paths = sorted(path_dir.glob("state_*"))

        if not paths:
            raise ValueError("No result files in dir " + str(path_dir))

        str_path = path = str(paths[0])
        if len(path) > 100:
            str_path = "[...]" + path[-100:]

        mpi.printby0("load params from file\n" + str_path)
        with h5py.File(path, "r") as h5file:
            return cls(hdf5_object=h5file["/info_simul/solver"])

    def __ior__(self, other):
        """Defines operator `|=`

        This operator performs a union with other Parameters instances.
        In simpler words, merge missing attributes and children (members).
        Do note, this does not check for mismatch in the parameter values.

        Parameters
        ----------
        other: Parameters
            Another Parameters instance to look for missing members.

        Returns
        -------
        Parameters

        """
        if not isinstance(other, type(self)):
            raise TypeError(
                f"{type(other)}. Can only merge instances of Parameters."
            )

        params1 = self
        params2 = other

        # Merge attributes
        diff_attribs = set(params2._key_attribs) - set(params1._key_attribs)

        if len(diff_attribs) > 0:
            print("Add parameter attributes: ", diff_attribs)

        for attrib in diff_attribs:
            params1._set_attrib(attrib, params2[attrib])

        # Merge children
        diff_children = set(params2._tag_children) - set(params1._tag_children)
        internal_attribs = [
            "attribs",
            "children",
            "key_attribs",
            "tag",
            "tag_children",
        ]

        if len(diff_children) > 0:
            print("Add parameter children: ", diff_children)

        for child in diff_children:
            child_attribs = params2[child]._make_dict()
            # Clean up internal attributes from dictionary
            list(map(child_attribs.__delitem__, internal_attribs))

            params1._set_child(child, child_attribs)

        # Recursive
        for child in params2._tag_children:
            params1[child] |= params2[child]

        return self



"""Utility to produce a dataframe from a set of simulations

"""

import json
import hashlib
import inspect
from abc import ABC, abstractmethod

from pathlib import Path

from pandas import DataFrame

from rich.progress import track


class DataframeMaker(ABC):
    """To produce a Pandas dataframe from a set of simulations"""

    @abstractmethod
    def get_time_start_from_path(self, path):
        """ "Get first time"""
        # t_start, _ = times_start_last_from_path(path)
        # return t_start

    @abstractmethod
    def get_time_last_from_path(self, path):
        """ "Get last saved time"""
        # return get_last_time_spatial_means_from_path(path)

    @abstractmethod
    def load_sim(self, path):
        """Load a simulation object"""
        # return load_sim_for_plot(path, hide_stdout=True)

    def get_mean_values_from_path(
        self, path, tmin=None, tmax=None, use_cache=True, customize=None
    ):
        """Get a dict of scalar values characterizing the simulation

        Parameters
        ----------

        tmin: float
            Minimum time

        tmax: float
            Maximum time

        use_cache: bool
            If True, return the cached result

        customize: callable

            If not None, called as ``customize(result, self.sim)`` to modify the
            returned dict.

        Examples
        --------

        .. code-block:: python

            def customize(result, sim):
                result["Rb"] = float(sim.params.short_name_type_run.split("_Rb")[-1])
            get_mean_values_from_path(path, customize=customize)

        """

        if (
            tmin is None
            or isinstance(tmin, str)
            or tmax is None
            or isinstance(tmax, str)
        ):
            t_start = self.get_time_start_from_path(path)
            t_last = self.get_time_last_from_path(path)

        if tmin is None:
            tmin = t_start
        elif isinstance(tmin, str):
            if tmin.startswith("t_start+"):
                tmin = t_start + float(tmin.split("t_start+")[-1])
            elif tmin.startswith("t_last-"):
                tmin = t_last - float(tmin.split("t_last-")[-1])
            else:
                raise ValueError(
                    f"isinstance(tmin, str) and {tmin=} but tmin has to start by "
                    '"t_start+" or "t_last-"'
                )
        tmin = float(tmin)

        if tmax is None:
            tmax = t_last
        elif isinstance(tmax, str):
            if tmax.startswith("t_start+"):
                tmax = t_start + float(tmax.split("t_start+")[-1])
            elif tmax.startswith("t_last-"):
                tmax = t_last - float(tmax.split("t_last-")[-1])
            else:
                raise ValueError(
                    f"isinstance(tmax, str) and {tmax=} but tmin has to start by "
                    '"t_start+" or "t_last-"'
                )
        tmax = float(tmax)

        cache_dir = Path(path) / ".cache"
        cache_dir.mkdir(exist_ok=True)

        if customize is not None:
            source = inspect.getsource(customize).encode().strip()
            hash = hashlib.sha256(source).hexdigest()[:16]
            part_customize = f"_customize{hash}"
        else:
            part_customize = ""

        cache_file = cache_dir / (
            f"mean_values_tmin{tmin}_tmax{tmax}{part_customize}.json"
        )

        if use_cache and cache_file.exists():
            with open(cache_file, "r") as file:
                return json.load(file)

        sim = self.load_sim(path)

        result = sim.output._compute_mean_values(tmin, tmax)
        if customize is not None:
            customize(result, sim)

        print("saving", cache_file)
        with open(cache_file, "w") as file:
            json.dump(result, file, indent=2)
        return result

    def get_dataframe_from_paths(
        self, paths, tmin=None, tmax=None, use_cache=True, customize=None
    ):
        """Produce a dataframe from a set of simulations.

        Uses ``sim.output.get_mean_values``

        """
        values = []
        for path in track(paths, "Getting the mean values"):
            values.append(
                self.get_mean_values_from_path(
                    path, tmin, tmax, use_cache, customize
                )
            )
        return DataFrame(values)



import sys
from shutil import copyfile

import pytest

import numpy as np

import pymech

name_solver = "testing"


@pytest.fixture
def path_sim(tmp_path):
    path_dir = tmp_path / "session_00"
    path_dir.mkdir(exist_ok=True, parents=True)

    nx = 2
    ny = 4
    nz = 6
    nx_elem = ny_elem = nz_elem = 2

    hexa_data = pymech.core.HexaData(
        ndim=3,
        nel=nx_elem * ny_elem * nz_elem,
        lr1=(nx, ny, nz),
        var=(3, 3, 1, 1, 2),
    )
    hexa_data.wdsz = 8
    hexa_data.istep = 0
    hexa_data.endian = sys.byteorder

    x1d = np.linspace(0, 1, nx)
    y1d = np.linspace(0, 1, ny)
    z1d = np.linspace(0, 1, nz)

    y3d, z3d, x3d = np.meshgrid(y1d, z1d, x1d)
    assert y3d.shape == (nz, ny, nx)

    ielem = 0
    for iz_elem in range(nz_elem):
        for iy_elem in range(ny_elem):
            for ix_elem in range(nx_elem):
                elem = hexa_data.elem[ielem]
                ielem += 1
                elem.pos[0] = x3d + ix_elem
                elem.pos[1] = y3d + iy_elem
                elem.pos[2] = z3d + iz_elem
                elem.vel.fill(1)

    time = 2.0
    for it in range(4):
        hexa_data.time = time
        path = path_dir / f"{name_solver}0.f{it:05d}"
        pymech.writenek(path, hexa_data)
        copyfile(path, path.with_name("sts" + path.name))
        time += 0.5

    return path_dir.parent


class Object:
    pass


@pytest.fixture
def false_output(path_sim):
    output = Object()
    output.name_solver = name_solver
    output.sim = Object()
    output.sim.params = Object()
    output.params = Object()
    output.sim.params.output = Object()
    output.sim.oper = Object()
    output.sim.oper.axes = "zyx"
    output.sim.params.output.session_id = 0
    output.path_run = path_sim
    return output



"""Utilities for fluidsim scripts

Provides:

.. autosummary::
   :toctree:

    restart

.. autofunction:: parse_args

"""

import shlex

from fluiddyn.util import mpi


def parse_args(parser, args=None):
    """Parse the arguments"""
    if args is not None and isinstance(args, str):
        args = shlex.split(args)
    args = parser.parse_args(args)
    mpi.printby0(args)
    return args



"""Concrete PhysField class for Snek5000
----------------------------------------

.. autoclass:: PhysFields4Snek5000
   :members:
   :private-members:
   :undoc-members:

"""

from functools import lru_cache, partial

from fluidsim_core.output.phys_fields import PhysFieldsABC
from fluidsim_core.output.movies import MoviesBasePhysFieldsHexa
from fluidsim_core.hexa_files import SetOfPhysFieldFiles


class PhysFields4Snek5000(PhysFieldsABC):
    _cls_movies = MoviesBasePhysFieldsHexa
    _cls_set_of_files = SetOfPhysFieldFiles

    def __init__(self, output=None):
        self.output = output
        self.params = output.params

        self.set_of_phys_files = self._cls_set_of_files(output=output)
        self.plot_hexa = self.set_of_phys_files.plot_hexa
        self.plot_hexa_stat = partial(
            self.set_of_phys_files.plot_hexa, prefix="sts"
        )
        self.read_hexadata = self.set_of_phys_files.read_hexadata
        self.read_hexadata_stat = partial(
            self.set_of_phys_files.read_hexadata, prefix="sts"
        )
        self.read_hexadata_from_time = (
            self.set_of_phys_files.read_hexadata_from_time
        )
        self.read_hexadata_from_time_stat = partial(
            self.set_of_phys_files.read_hexadata_from_time, prefix="sts"
        )

        self.movies = self._cls_movies(output, self)
        self.animate = self.movies.animate
        self.interact = self.movies.interact
        self._equation = "z=0"

    def get_key_field_to_plot(self, forbid_compute=False, key_prefered=None):
        return self.set_of_phys_files.get_key_field_to_plot(key_prefered)

    def get_field_to_plot(
        self,
        key=None,
        time=None,
        idx_time=None,
        equation=None,
        interpolate_time=True,
        skip_vars=(),
    ):
        """Get the field to be plotted in process 0."""
        if equation is None:
            equation = self._equation

        return self.set_of_phys_files.get_field_to_plot(
            time=time,
            idx_time=idx_time,
            key=key,
            equation=equation,
            interpolate_time=interpolate_time,
            skip_vars=skip_vars,
        )

    def get_vector_for_plot(
        self,
        from_state=False,
        time=None,
        interpolate_time=True,
        skip_vars=(),
    ):
        if from_state:
            raise ValueError("cannot get anything from the state for this solver")
        return self.set_of_phys_files.get_vector_for_plot(
            time, self._equation, skip_vars=skip_vars
        )

    @lru_cache(maxsize=None)
    def _get_axis_data(self, equation):
        (
            letter_x_axis,
            letter_y_axis,
        ) = self.set_of_phys_files.get_letters_axes_from_equation(equation)

        hexa_x, _ = self.get_field_to_plot(
            idx_time=0, key=letter_x_axis, equation=equation
        )
        hexa_y, _ = self.get_field_to_plot(
            idx_time=0, key=letter_y_axis, equation=equation
        )
        return hexa_x, hexa_y



"""Output API

.. autoclass:: SimReprMakerCore
   :members:
   :private-members:

.. autoclass:: OutputCore
   :members:
   :private-members:
   :special-members: __init__

"""

from abc import ABC, abstractmethod, abstractstaticmethod
from pathlib import Path
from time import sleep

import fluiddyn
from fluiddyn.io import FLUIDDYN_PATH_SCRATCH, FLUIDSIM_PATH
from fluiddyn.util import time_as_str, mpi

from fluidsim_core import __version__


class SimReprMakerCore:
    """Code to create strings to represent a simulation"""

    def __init__(self, sim):
        self.sim = sim
        self.params = sim.params
        self.ordered_keys = []
        self.parameters = {}
        self.formats = {}

    def add_word(self, word):
        if word:
            self.ordered_keys.append(("__word", word))

    def add_parameters(self, parameters, formats=None, indices=None):
        for key in parameters:
            if indices and key in indices:
                index = indices[key]
                self.ordered_keys.insert(index, ("__parameter", key))
            else:
                self.ordered_keys.append(("__parameter", key))
        self.parameters.update(parameters)
        if formats:
            self.formats.update(formats)

    def _make_list_repr(self):
        list_repr = []
        for kind, value in self.ordered_keys:
            if kind == "__word":
                list_repr.append(value)
            elif kind == "__parameter":
                name_parameter = value
                parameter = self.parameters[name_parameter]
                if isinstance(parameter, float):
                    default_fmt = ".3f"
                else:
                    default_fmt = ""
                fmt = self.formats.get(name_parameter, default_fmt)
                str_parameter = ("{:" + fmt + "}").format(parameter)
                if fmt and fmt[-1] in ["e", "g"] and "e+" in str_parameter:
                    str_parameter = str_parameter.replace("e+", "e")
                if "e" not in str_parameter and "." in str_parameter:
                    str_parameter = str_parameter.rstrip("0")
                if str_parameter.endswith("."):
                    str_parameter = str_parameter[:-1]
                list_repr.append((name_parameter, str_parameter))
            else:
                raise ValueError
        return list_repr

    def get_list_repr(self):
        if not hasattr(self, "_list_repr"):
            self._list_repr = self._make_list_repr()
        return self._list_repr

    def time_from_path_run(self, path_run):
        if isinstance(path_run, Path):
            path_run = path_run.name
        return "_".join(path_run.split("_")[-2:])

    def get_time_as_str(self):
        params = self.sim.params
        if not params.NEW_DIR_RESULTS:
            return self.time_from_path_run(params.path_run)
        else:
            return time_as_str()

    def make_representations(self):
        """Generate a unique name and summary for the simulation run"""
        time = self.get_time_as_str()
        list_repr = self.get_list_repr().copy()
        list_repr.append(time)

        for_name = []
        for_summary = []
        for obj in list_repr:
            if isinstance(obj, tuple):
                name_parameter, str_parameter = obj
                obj = f"{name_parameter}{str_parameter}"
                str_summary = f"{name_parameter}={str_parameter}"
            else:
                str_summary = obj.replace("_", ", ")

            for_name.append(obj)
            for_summary.append(str_summary)

        name_run = "_".join(for_name)
        summary_simul = ", ".join(for_summary)
        return name_run, summary_simul


class OutputCore(ABC):
    """Base Output class"""

    SimReprMaker = SimReprMakerCore

    @abstractstaticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver. More *specialized output*
        classes may be assigned as follows:

        .. code-block:: python

            classes = info_solver.classes.Output.classes
            classes._set_child(
                "PrintStdOut",
                attribs={
                    "module_name": "package.module",
                    "class_name": "MyPrintStdOut",
                },
            )

        """
        info_solver.classes.Output._set_child("classes")

    @abstractstaticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container."""
        attribs = {
            "HAS_TO_SAVE": True,
            "sub_directory": "",
            # Possibly more ...
        }
        params._set_child("output", attribs=attribs)
        # params.output._set_doc("...")

    @abstractmethod
    def __init__(self, sim):
        """Initializes the instance attributes and child *output* classes.

        .. note::
            This class relies on a boolean parameter ``params.NEW_DIR_RESULTS``.

        """
        #: The simulation class
        self.sim = sim

        if hasattr(sim, "oper"):
            #: An alias towards the Operators object
            self.oper = sim.oper

        params = sim.params
        #: The tree of parameters for output-related classes
        self.params = params.output

        #: Determines whether to save output files on-the-fly
        self._has_to_save = self.params.HAS_TO_SAVE
        #: Alias for solver short-name
        self.name_solver = sim.info.solver.short_name

        # initialisation name_run and path_run
        if mpi.rank == 0:
            self._init_name_run()
        else:
            self.name_run = None

        if not params.NEW_DIR_RESULTS:
            try:
                self.path_run = str(params.path_run)
            except AttributeError:
                params.NEW_DIR_RESULTS = True
                print(
                    "Strange: params.NEW_DIR_RESULTS == False "
                    "but no params.path_run"
                )

            # if _has_to_save, we verify the correspondence between the
            # resolution of the simulation and the resolution of the
            # previous simulation saved in this directory
            if self._has_to_save:
                if mpi.rank == 0:
                    try:
                        params_dir = sim.Parameters._load_params_simul(
                            self.path_run
                        )
                    except IOError:
                        raise ValueError(
                            "Strange, no info_simul.h5 in self.path_run"
                        )

                    cond = False
                    for n in ("nx", "ny", "nz"):
                        try:
                            # Mismatch in resolution
                            if getattr(params.oper, n) != getattr(
                                params_dir.oper, n
                            ):
                                cond = True
                                break
                        except AttributeError:
                            pass

                    if cond:
                        params.NEW_DIR_RESULTS = True
                        print(
                            """
Warning: params.NEW_DIR_RESULTS is False but the resolutions of the simulation
         and of the simulation in the directory self.path_run are different
         we put params.NEW_DIR_RESULTS = True"""
                        )
                if mpi.nb_proc > 1:
                    params.NEW_DIR_RESULTS = mpi.comm.bcast(
                        params.NEW_DIR_RESULTS
                    )

        if params.NEW_DIR_RESULTS:
            #: Full path of the output directory for this specific simulation
            self.path_run = self._init_path_run()

        if mpi.nb_proc > 1:
            # ensure same name_run across all processes
            self.name_run = mpi.comm.bcast(self.name_run, root=0)

        self.sim.name_run = self.name_run

        #: Optionally an instance of class PrintStdout which writes output
        #  to stdout and also a file buffer
        self.print_stdout = print

    def _init_path_run(self):
        """Initialize a unique path for the simulation.

        Returns
        -------
        str

        """
        params = self.sim.params

        if FLUIDDYN_PATH_SCRATCH is not None:
            path_base = FLUIDDYN_PATH_SCRATCH
        else:
            path_base = FLUIDSIM_PATH

        path_base = Path(path_base)

        if len(params.output.sub_directory) > 0:
            path_base = path_base / params.output.sub_directory

        if mpi.rank == 0:
            while True:
                path_run = path_base / self.name_run
                if not params.output.HAS_TO_SAVE:
                    break
                if not path_run.exists():
                    try:
                        path_run.mkdir(parents=True)
                    except OSError:
                        # in case of simultaneously launched simulations
                        print(
                            'Warning: NEW_DIR_RESULTS=True, but path"',
                            path_run,
                            "already exists. Trying a new path...",
                        )
                        sleep(1)
                        self._init_name_run()
                    else:
                        break
                else:
                    sleep(1)
                    self._init_name_run()

            path_run = str(path_run)
        else:
            path_run = None

        if mpi.rank == 0:
            params._set_attrib("path_run", path_run)
        if mpi.nb_proc > 1:
            path_run = mpi.comm.bcast(path_run, root=0)

        return path_run

    def _init_sim_repr_maker(self):
        """Create a list of strings to make the run name.

        Returns
        -------
        :any:`fluidsim_core.output.base.SimReprMakerCore`

        """
        sim = self.sim

        sim_repr_maker = self.SimReprMaker(sim)
        sim_repr_maker.add_word(self.name_solver)
        sim_repr_maker.add_word(sim.params.short_name_type_run)

        # Modify sim_repr_maker as necessary
        return sim_repr_maker

    def _init_name_run(self):
        """Initialize the ``name_run`` and ``summary_simul`` attributes by
        calling :any:`fluidsim_core.output.base.SimReprMakerCore.make_representations`

        """
        if not hasattr(self, "_sim_repr_maker"):
            self._sim_repr_maker = self._init_sim_repr_maker()
        (
            #: Name of the output directory for this specific simulation
            self.name_run,
            self.summary_simul,
        ) = self._sim_repr_maker.make_representations()

    @abstractmethod
    def post_init(self):
        """Execute once the sim object is injected with all child classes.
        Typically used to print descriptive initialization messages.

        """
        sim = self.sim

        if mpi.rank == 0:
            objects_to_print = {
                "sim": sim,
                "sim.output": sim.output,
            }

            for key, obj in objects_to_print.items():
                self.print_stdout(f"{key + ': ':20s}" + str(obj.__class__))
        self._save_info_solver_params_xml()

    def _save_info_solver_params_xml(self, replace=False, comment=""):
        """Save files with information on the solver and on the run."""
        if (
            mpi.rank == 0
            and self._has_to_save
            and self.sim.params.NEW_DIR_RESULTS
        ):
            comment = f"""\
This file should not be modified (except for adding xml comments).
Created by the Python programs:
FluidDyn {fluiddyn.__version__}
FluidSim Core {__version__}
{comment}
"""

            path_run = Path(self.path_run)
            info_solver_xml_path = path_run / "info_solver.xml"
            params_xml_path = path_run / "params_simul.xml"

            # save info on the run
            if replace:
                info_solver_xml_path.unlink()
                params_xml_path.unlink()

            self.sim.info.solver._save_as_xml(
                path_file=info_solver_xml_path, comment=comment
            )

            self.sim.params._save_as_xml(
                path_file=params_xml_path, comment=comment
            )



"""Abstract base classes for Physical fields output
---------------------------------------------------

These classes can be used to define classes working together with
:class:`fluidsim_core.output.movies.MoviesBasePhysFields`.

.. autoclass:: PhysFieldsABC
   :members:
   :private-members:
   :undoc-members:

.. autoclass:: SetOfPhysFieldFilesABC
   :members:
   :private-members:
   :undoc-members:

"""

from abc import ABCMeta, abstractmethod
from typing import Optional
from pathlib import Path
import math

import numpy as np


class SetOfPhysFieldFilesABC(metaclass=ABCMeta):
    times: list

    @abstractmethod
    def update_times(self):
        "Update the times of the files"

    @abstractmethod
    def get_min_time(self):
        "Get minimum time"


class PhysFieldsABC(metaclass=ABCMeta):
    _equation: Optional[str]
    _can_plot_quiver: Optional[bool]
    set_of_phys_files: SetOfPhysFieldFilesABC

    @abstractmethod
    def get_key_field_to_plot(self, forbid_compute=False, key_prefered=None):
        "Get the key corresponding to the field to be plotted"

    @abstractmethod
    def get_field_to_plot(
        self,
        key=None,
        time=None,
        idx_time=None,
        equation=None,
        interpolate_time=True,
        skip_vars=(),
    ):
        """Get the field to be plotted in process 0."""

    @abstractmethod
    def get_vector_for_plot(
        self, from_state=False, time=None, interpolate_time=True
    ):
        "Get the vector components"

    def _set_title(self, ax, key, time, vmax=None):
        title = f"{key}, $t = {time:.3f}$"
        if vmax is not None:
            title += r", $|\vec{v}|_{max} = $" + f"{vmax:.3f}"
        ax.set_title(title)

    @abstractmethod
    def _get_axis_data(equation=None):
        "Get x and y axes data"

    def set_equation_crosssection(self, equation):
        """Set the equation defining the cross-section.

        Parameters
        ----------

        equation : str

          The equation can be of the shape 'iz=2', 'z=1', ...

        """
        self._equation = equation

    def _init_skip_quiver(self):
        params = self.output.sim.params
        Lx = params.oper.Lx
        Ly = params.oper.Ly
        x, y = self._get_axis_data()
        # 4% of the Lx it is a great separation between vector arrows.
        try:
            delta_quiver = 0.04 * min(Lx, Ly)
        except AttributeError:
            skip = 1
        else:
            skip = (len(x) / Lx) * delta_quiver
            skip = int(np.round(skip))
            if skip < 1:
                skip = 1
        self._skip_quiver = skip
        return skip


class SetOfPhysFieldFilesBase(SetOfPhysFieldFilesABC):
    @staticmethod
    @abstractmethod
    def time_from_path(path):
        "Get the time corresponding to a file"

    def __init__(self, path_dir=None, output=None):
        self.output = output
        if path_dir is None:
            if output is None:
                path_dir = Path.cwd()
            else:
                path_dir = Path(output.path_run)
        self.path_dir = Path(path_dir)
        self._glob_pattern = self._get_glob_pattern()
        self.update_times()

    def _get_glob_pattern(self):
        return "state_phys*.[hn]*"

    def update_times(self):
        """Initialize the times by globing and analyzing the file names."""
        path_files = sorted(self.path_dir.glob(self._glob_pattern))

        if hasattr(self, "path_files") and len(self.path_files) == len(
            path_files
        ):
            return

        self.path_files = sorted(path_files)
        self.times = np.array(
            [self.time_from_path(path) for path in self.path_files]
        )

    def get_min_time(self):
        if hasattr(self, "times"):
            return self.times.min()
        return 0.0

    def get_max_time(self):
        return self.times.max()

    def get_field_to_plot(
        self,
        time=None,
        idx_time=None,
        key=None,
        equation=None,
        interpolate_time=True,
        skip_vars=(),
    ):
        if time is None and idx_time is None:
            self.update_times()
            time = self.times[-1]

        # Assert files are available
        if self.times.size == 0:
            raise FileNotFoundError(
                f"No phys files were detected in directory: {self.path_dir}"
            )

        if not interpolate_time and time is not None:
            idx, time_closest = self.get_closest_time_file(time)
            return self.get_field_to_plot(
                idx_time=idx, key=key, equation=equation, skip_vars=skip_vars
            )

        if interpolate_time and time is not None:
            idx_closest, time_closest = self.get_closest_time_file(time)

            if math.isclose(time, time_closest) or self.times.size == 1:
                return self.get_field_to_plot(
                    idx_time=idx_closest,
                    key=key,
                    equation=equation,
                    skip_vars=skip_vars,
                )

            if idx_closest == self.times.size - 1:
                idx0 = idx_closest - 1
                idx1 = idx_closest
            elif time_closest < time or idx_closest == 0:
                idx0 = idx_closest
                idx1 = idx_closest + 1
            elif time_closest > time:
                idx0 = idx_closest - 1
                idx1 = idx_closest

            dt_save = self.times[idx1] - self.times[idx0]
            weight0 = 1 - np.abs(time - self.times[idx0]) / dt_save
            weight1 = 1 - np.abs(time - self.times[idx1]) / dt_save

            field0, time0 = self.get_field_to_plot(
                idx_time=idx0, key=key, equation=equation, skip_vars=skip_vars
            )
            field1, time1 = self.get_field_to_plot(
                idx_time=idx1, key=key, equation=equation, skip_vars=skip_vars
            )

            return field0 * weight0 + field1 * weight1, time

        return self._get_field_to_plot_from_file(
            self.path_files[idx_time], key, equation=equation, skip_vars=skip_vars
        )

    def get_closest_time_file(self, time):
        """Find the index and value of the closest actual time of the field."""
        idx = np.abs(self.times - time).argmin()
        return idx, self.times[idx]

    @abstractmethod
    def _get_field_to_plot_from_file(
        self, path_file, key, equation, skip_vars=()
    ):
        "Get a 2d field from a file"



from fluidsim_core.info import InfoSolverCore
from fluidsim_core.output import OutputCore
from fluidsim_core.params import Parameters, iter_complete_params
from fluidsim_core.tests.test_solver import SimulTest

import pytest


class InfoSolverTestWithOutput(InfoSolverCore):
    def _init_root(self):
        super()._init_root()

        self._set_attribs(
            {
                "module_name": "fluidsim_core.tests.test_solver_with_output",
                "class_name": "SimulTestWithOutput",
                "short_name": "TestWithOutput",
            }
        )
        classes = self.classes
        classes._set_child(
            "Output",
            attribs={
                "module_name": "fluidsim_core.tests.test_solver_with_output",
                "class_name": "OutputTest",
            },
        )


class SimulTestWithOutput(SimulTest):
    InfoSolver = InfoSolverTestWithOutput

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        params._set_attribs(
            {"NEW_DIR_RESULTS": True, "short_name_type_run": "test"}
        )

    def __init__(self, params):
        super().__init__(params)
        dict_classes = self.info_solver.import_classes()

        for cls_name, Class in dict_classes.items():
            # only initialize if Class is not the Simul class
            if not isinstance(self, Class):
                setattr(self, cls_name.lower(), Class(self))

        self.output.post_init()


class OutputTest(OutputCore):
    @staticmethod
    def _complete_info_solver(info_solver):
        OutputCore._complete_info_solver(info_solver)
        classes = info_solver.classes.Output.classes

        # NOTE: This is wrong and added to check if a corresponding UserWarning
        # is invoked
        classes._set_attrib(
            "Dummy",
            dict(
                module_name="fluidsim_core.tests.test_solver_with_output",
                class_name="DummyOutput",
            ),
        )

        # NOTE: This is correct :)
        classes._set_child(
            "Working",
            dict(
                module_name="fluidsim_core.tests.test_solver_with_output",
                class_name="WorkingOutput",
            ),
        )

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        OutputCore._complete_params_with_default(params, info_solver)
        dict_classes = info_solver.classes.Output.import_classes()
        iter_complete_params(params, info_solver, dict_classes.values())

    def __init__(self, sim):
        super().__init__(sim)

        dict_classes = self.sim.info.solver.classes.Output.import_classes()
        # initialize objects
        for cls_name, Class in dict_classes.items():
            # only initialize if Class is not the Output class
            if not isinstance(self, Class):
                setattr(self, cls_name.lower(), Class(self))

    def post_init(self):
        super().post_init()


class SpecificOutputTest:
    def __init__(self, sim):
        pass


class DummyOutput(SpecificOutputTest):
    """This class is not specified in the InfoSolver instance
    properly (see ``_complete_info_solver`` method above) and thus won't
    be instantiated.

    """


class WorkingOutput(SpecificOutputTest):
    @staticmethod
    def _complete_params_with_default(params):
        params.output._set_child("working", attribs={"works": 42})


def assert_default_params(params):
    assert not hasattr(params.output, "dummy")
    assert hasattr(params.output, "working")
    assert params.output.working.works == 42


def test_init_info_solver_params():
    with pytest.warns(UserWarning, match=r"A class Dummy .* using _set_attrib"):
        info_solver = InfoSolverTestWithOutput()

    params = Parameters._create_params(info_solver)

    assert_default_params(params)


def test_init_sim_output():
    with pytest.warns(UserWarning, match=r"A class Dummy .* using _set_attrib"):
        params = SimulTestWithOutput.create_default_params()

    assert_default_params(params)
    params.output.sub_directory = "tests"
    sim = SimulTestWithOutput(params)

    assert hasattr(sim.output, "working")



"""Outputs

.. autosummary::
   :toctree:

   base
   movies
   phys_fields
   phys_fields_snek5000
   remaining_clock_time

"""

from .base import SimReprMakerCore, OutputCore

__all__ = ["SimReprMakerCore", "OutputCore"]



from fluidsim_core.tests.solver import SimulTest


def test_init_sim():
    params = SimulTest.create_default_params()
    assert params.foo == 42
    params.foo = 24
    sim = SimulTest(params)
    assert sim.params.foo == 24
